.msp430

.include "msp430x2xx.inc"
.include "waveshare.inc"

LCD_RESET equ 0x02
LCD_BUSY equ 0x04
LCD_DC equ 0x08
SPI_CS equ 0x10
SPI_CLK equ 0x20
SPI_SDO equ 0x40

WIDTH equ 200
HEIGHT equ 200

RAM equ 0x200
DIGIT_0 equ RAM+0
DIGIT_1 equ RAM+2
DIGIT_2 equ RAM+4
DS18B20_DATA equ 0x0210
DS18B20_CMD equ 0x0220
DS18B20_CMD_LEN equ 0x0221
DS18B20_BIT equ 0x0222
DEBUG_STATE equ 0x0223
DS18B20_SERIAL equ 0x0230
DEBUG_DIGIT_0 equ 0x240
DEBUG_DIGIT_1 equ 0x242

.macro lcd_command(value)
  mov.b #value, r15
  call #send_command
.endm

.macro lcd_data(value)
  mov.b #value, r15
  call #send_data
.endm

  org 0xf800
start:
  ;; Turn off watchdog
  mov.w #0x5a80, &WDTCTL

  ;; Turn off interrupts
  dint

  ;; Set up stack pointer
  mov.w #0x280, SP

  ;; Set MCLK to 8 MHz with DCO 
  mov.b #DCO_5, &DCOCTL
  mov.b #RSEL_13, &BCSCTL1
  mov.b #0, &BCSCTL2

  ;; Set up output pins
  ;; P1.0=DS18B20 thermometer
  ;; P1.1=LCD_RESET    0x02
  ;; P1.2=LCD_BUSY     0x04
  ;; P1.3=LCD_DC       0x08
  ;; P1.4=SPI_CS       0x10
  ;; P1.5=SPI_CLK      0x20
  ;; P1.6=SPI_SDO      0x40
  mov.b #LCD_RESET|LCD_DC|SPI_CS|1, &P1DIR
  mov.b #SPI_CS|1, &P1OUT

  ;; Set up SPI
  mov.b #USIPE6|USIPE5|USIMST|USIOE|USISWRST, &USICTL0
  mov.b #USICKPH, &USICTL1
  mov.b #USIDIV_0|USISSEL_2, &USICKCTL
  bic.b #USISWRST, &USICTL0      ; clear reset

  ;; Set up Timer
  mov.w #2000, &TACCR0
  mov.w #(TASSEL_2|MC_1), &TACTL ; SMCLK, DIV1, COUNT to TACCR0
  mov.w #0, &TACCTL0
  mov.w #0, &TACCTL1

  ;; Turn on interrupts
  eint

  call #init_lcd
  call #set_memory_area
  call #clear_screen
  call #display_frame
  call #set_memory_area
  call #clear_screen
  call #display_frame

  mov.w #digits, &DIGIT_0
  mov.w #digits, &DIGIT_1
  mov.w #digits, &DIGIT_2

  call #set_memory_area
  call #draw_digits
  call #display_frame

  call #read_serial_number

main:
  ;; Set MCLK to 8 MHz with DCO 
  mov.b #DCO_5, &DCOCTL
  mov.b #RSEL_13, &BCSCTL1
  mov.b #0, &BCSCTL2

  mov.w #digits+8, &DIGIT_0
  mov.w #digits+8, &DIGIT_1
  mov.w #digits+8, &DIGIT_2

  call #read_temp

  ;; Compute whole number values
  mov.w &DS18B20_DATA, r15
  rra.w r15
  rra.w r15
  rra.w r15
  rra.w r15
  mov.w #0, r14
divide_value:
  cmp.w #10, r15
  jlo stop_dividing
  sub.w #10, r15
  add.w #1, r14
  jmp divide_value
stop_dividing:
  mov.w r15, &DEBUG_DIGIT_1
  mov.w r14, &DEBUG_DIGIT_0
  rla.w r15
  rla.w r15
  rla.w r15
  rla.w r14
  rla.w r14
  rla.w r14
  add #digits, r15
  add #digits, r14
  mov.w r15, &DIGIT_1
  mov.w r14, &DIGIT_0

  ;; Compute decimal value
  mov.w &DS18B20_DATA, r15
  and.w #0x000f, r15
  add.w #dec_value, r15
  mov.b @r15, r15
  rla.w r15
  rla.w r15
  rla.w r15
  add.w #digits, r15
  mov.w r15, &DIGIT_2

  call #set_memory_area
  call #draw_digits
  call #display_frame

  ;; Set MCLK to 96,000 KHz with DCO (lower power consumption)
  mov.b #DCO_1, &DCOCTL
  mov.b #RSEL_0, &BCSCTL1
  mov.b #0, &BCSCTL2

  ;; Count out approximately 1 minute
  mov.w #30, r14
main_delay_loop_outer:
  mov.w #0xffff, r15
main_delay_loop:
  dec r15
  jnz main_delay_loop
  dec r14
  jnz main_delay_loop_outer
  jmp main

init_lcd:
  ;; Reset LCD
  bic.b #LCD_RESET, &P1OUT
  call #delay
  bis.b #LCD_RESET, &P1OUT
  call #delay
  call #delay

  lcd_command(DRIVER_OUTPUT_CONTROL)
  ;lcd_data((HEIGHT - 1) & 0xff)
  ;lcd_data(((HEIGHT - 1) >> 8) & 0xff)
  lcd_data(199)
  lcd_data(0)
  lcd_data(0x00)
  lcd_command(BOOSTER_SOFT_START_CONTROL)
  lcd_data(0xd7)
  lcd_data(0xd6)
  lcd_data(0x9d)
  lcd_command(WRITE_VCOM_REGISTER)
  lcd_data(0xa8)
  lcd_command(SET_DUMMY_LINE_PERIOD)
  lcd_data(0x1a)
  lcd_command(SET_GATE_TIME)
  lcd_data(0x08)
  lcd_command(DATA_ENTRY_MODE_SETTING)
  lcd_data(0x03)

  lcd_command(WRITE_LUT_REGISTER)
  mov.w #30, r4
  mov.w #lut_full_update, r5
  ;mov.w #lut_partial_update, r5
send_lut:
  mov.b @r5+, r15
  call #send_data
  dec.w r4
  jnz send_lut

  ret

set_memory_area:
  ;; Set memory area
  lcd_command(SET_RAM_X_ADDRESS_START_END_POSITION)
  lcd_data(0)
  lcd_data(199 >> 3)
  lcd_command(SET_RAM_Y_ADDRESS_START_END_POSITION)
  lcd_data(0)
  lcd_data(0)
  lcd_data(199)
  lcd_data(0)

  ;; Set memory pointer
  lcd_command(SET_RAM_X_ADDRESS_COUNTER)
  lcd_data(0)
  lcd_command(SET_RAM_Y_ADDRESS_COUNTER)
  lcd_data(0)
  lcd_data(0)
  lcd_data(0)

  call #wait_busy
  ret

clear_screen:
  lcd_command(WRITE_RAM)
  ;; (200 / 8) * 200 bytes
  mov.w #5000, r4
send_clear_byte:
  lcd_data(0xff)
  dec.w r4
  jnz send_clear_byte
  call #wait_busy
  ret

write_digit_piece:
  mov.w #8, r13
write_next_8_pixels:
  mov.w #0x00, r15
  add.b r14, r14
  jc dont_modify_pixels
  mov.w #0xff, r15 
dont_modify_pixels:
  call #send_data
  dec.w r13
  jnz write_next_8_pixels
  ret

write_digit_piece_with_dot:
  mov.w #7, r13
write_next_7_pixels:
  mov.w #0x00, r15
  add.b r14, r14
  jc dont_modify_pixels_with_dot
  mov.w #0xff, r15 
dont_modify_pixels_with_dot:
  call #send_data
  dec.w r13
  jnz write_next_7_pixels
  mov.b #0x81, r15 
  call #send_data
  ret

draw_digits:
  lcd_command(WRITE_RAM)

  ;; Clear top 68 lines of display
  ;; (200 / 8) * 68 bytes
  mov.w #1700, r4
send_top_byte:
  lcd_data(0xff)
  dec.w r4
  jnz send_top_byte

  ;; Draw the 3 digits and period in the next 64 lines.
  mov.w #8, r4
send_next_digit_block:
  mov.w #8, r5
send_next_digit_line:
  ;; Left column needs to be cleared.
  lcd_data(0xff)
  mov.w &DIGIT_0, r14
  mov.b @r14, r14
  call #write_digit_piece
  mov.w &DIGIT_1, r14
  mov.b @r14, r14
  cmp.w #1, r4
  jnz display_normal
  call #write_digit_piece_with_dot
  jmp display_normal_finish
display_normal:
  call #write_digit_piece
display_normal_finish:
  mov.w &DIGIT_2, r14
  mov.b @r14, r14
  call #write_digit_piece
  dec.w r5
  jnz send_next_digit_line
  add.w #1, &DIGIT_0
  add.w #1, &DIGIT_1
  add.w #1, &DIGIT_2
  dec.w r4
  jnz send_next_digit_block

  ;; Clear bottom 68 lines of display
  ;; (200 / 8) * 68 bytes
  mov.w #1700, r4
send_bottom_byte:
  lcd_data(0xff)
  dec.w r4
  jnz send_bottom_byte

  call #wait_busy
  ret

display_frame:
  ;; Update the display
  lcd_command(DISPLAY_UPDATE_CONTROL_2)
  lcd_data(0xc4)
  lcd_command(MASTER_ACTIVATION)
  lcd_command(TERMINATE_FRAME_READ_WRITE)
  call #wait_busy
  ret

;; send_command(r15)
send_command:
  bic.b #LCD_DC, &P1OUT          ; /DC=command
  bic.b #SPI_CS, &P1OUT          ; /CS=selected

  mov.b r15, &USISRL
  mov.b #8, &USICNT
send_command_wait_spi:
  bit.b #USIIFG, &USICTL1
  jz send_command_wait_spi

  bis.b #SPI_CS, &P1OUT          ; /CS=deselected
  ret

;; send_data(r15)
send_data:
  bis.b #LCD_DC, &P1OUT          ; /DC=data
  bic.b #SPI_CS, &P1OUT          ; /CS=selected

  mov.b r15, &USISRL
  mov.b #8, &USICNT
send_data_wait_spi:
  bit.b #USIIFG, &USICTL1
  jz send_data_wait_spi

  bis.b #SPI_CS, &P1OUT          ; /CS=deselected
  ret

wait_busy:
  bit.b #LCD_BUSY, &P1IN
  jnz wait_busy
  ret

deep_sleep:
  lcd_command(DEEP_SLEEP_MODE)
  call #wait_busy
  ret

delay:
  mov.w #8, r14
delay_loop_outer:
  mov.w #0xffff, r15
delay_loop:
  dec r15
  jnz delay_loop
  dec r14
  jnz delay_loop_outer
  ret

read_serial_number:
  mov.w #CCIE, &TACCTL0
  call #init_ds18b20
  mov.b #0x33, &DS18B20_CMD
  call #send_ds18b20_cmd
  call #fetch_ds18b20_serial
  mov.w #0, &TACCTL0
  ret

read_temp:
  mov.w #CCIE, &TACCTL0
  ;; First do a convert
  call #init_ds18b20
  mov.b #0xcc, &DS18B20_CMD
  call #send_ds18b20_cmd
  mov.b #0x44, &DS18B20_CMD
  call #send_ds18b20_cmd
  ;;mov.w #0, r4  ;; DEBUG
  call #wait_ds18b20_convert

  ;; Second fetch the scratch-pad
  call #init_ds18b20
  mov.b #0xcc, &DS18B20_CMD
  call #send_ds18b20_cmd
  mov.b #0xbe, &DS18B20_CMD
  call #send_ds18b20_cmd
  call #fetch_ds18b20_scratch_pad

  mov.w #0, &TACCTL0
  ret

;; init DS18B20
init_ds18b20:
  bic.b #1, &P1OUT
  mov.w #3840, &TACCR0   ; 480us
  mov.w #0, &TAR
  mov.w #0, r11
master_reset_pulse:
  cmp.w #1, r11
  jne master_reset_pulse
  bis.b #1, &P1OUT
  bic.b #1, &P1DIR
wait_slave_low:
  bit.b #1, &P1IN
  jnz wait_slave_low
wait_slave_hi:
  bit.b #1, &P1IN
  jz wait_slave_hi
  bis.b #1, &P1DIR
  mov.w #480, &TACCR0   ; 60us
  ret

;; send_ds18b20_cmd
send_ds18b20_cmd:
  mov.b #8, &DS18B20_CMD_LEN
send_next_bit:
  bit.b #1, &DS18B20_CMD
  jz cmd_bit_0
  call #write_ds18b20_1
  jmp cmd_bit_done
cmd_bit_0:
  call #write_ds18b20_0
cmd_bit_done:
  mov.w #6, r10
cmd_wait_1us:
  dec r10
  jnz cmd_wait_1us
  rra.b &DS18B20_CMD
  dec.b &DS18B20_CMD_LEN
  jnz send_next_bit
  ret

;; wait_ds18b20_convert
wait_ds18b20_convert:
  mov.w #6, r10
convert_wait_1us:
  dec r10
  jnz convert_wait_1us
  call #read_ds18b20
  cmp.w #0, r10
  jeq wait_ds18b20_convert
  ret

;; fetch_ds18b20_scratch_pad
fetch_ds18b20_scratch_pad:
  mov.w #DS18B20_DATA, r14
fetch_next_byte:
  mov.w #8, &DS18B20_BIT

fetch_next_bit:
  mov.w #6, r10
scratch_wait_1us:
  dec r10
  jnz scratch_wait_1us

  call #read_ds18b20
  rra.b 0(r14)
  cmp.w #0, r10
  jeq fetched_0
  bis.b #128, 0(r14)
  jmp done_set_fetch
fetched_0:
  bic.b #128, 0(r14)
done_set_fetch:
  dec &DS18B20_BIT
  jnz fetch_next_bit

  inc r14
  cmp.w #DS18B20_DATA+9, r14
  jne fetch_next_byte
  ret

;; write 1 to DS18B20
write_ds18b20_1:
  bic.b #1, &P1OUT
  mov.w #0, &TAR
  mov.w #0, r11
  mov.w #6, r10
wait_at_least_1us_write_1:
  dec r10
  jnz wait_at_least_1us_write_1
  bis.b #1, &P1OUT
master_write_1:
  cmp.w #1, r11
  jne master_write_1
  ret

;; write 0 to DS18B20
write_ds18b20_0:
  bic.b #1, &P1OUT
  mov.w #0, &TAR
  mov.w #0, r11
master_write_0:
  cmp.w #1, r11
  jne master_write_0
  bis.b #1, &P1OUT
  ret

;; read from DS18B20 (returns 0 in r10 for a 0, return 1 for a 1)
read_ds18b20:
  bic.b #1, &P1OUT
  mov.w #6, r10
wait_at_least_1us_read:
  dec r10
  jnz wait_at_least_1us_read
  bis.b #1, &P1OUT
  bic.b #1, &P1DIR
  mov.w #0, &TAR
  mov.w #0, r11
wait_read_bit:
  bit.w #1, &P1IN
  jnz read_bit_not_zero
  inc r10
read_bit_not_zero:
  cmp.w #1, r11
  jne wait_read_bit
  bis.b #1, &P1DIR

  cmp.w #0, r10
  jeq return_1
  mov.w #0, r10
  ret
return_1:
  mov.w #1, r10
  ret

;; fetch_ds18b20_serial
fetch_ds18b20_serial:
  mov.w #DS18B20_SERIAL, r14
serial_next_byte:
  mov.w #8, &DS18B20_BIT

serial_next_bit:
  mov.w #6, r10
serial_wait_1us:
  dec r10
  jnz serial_wait_1us

  call #read_ds18b20
  rra.b 0(r14)
  cmp.w #0, r10
  jeq fetched_serial_0
  bis.b #128, 0(r14)
  jmp done_set_fetch_serial
fetched_serial_0:
  bic.b #128, 0(r14)
done_set_fetch_serial:
  dec &DS18B20_BIT
  jnz serial_next_bit

  inc r14
  cmp.w #DS18B20_SERIAL+8, r14
  jne serial_next_byte
  ret


timer_interrupt:
  mov.w #1, r11
  reti

lut_full_update:
  db 0x02, 0x02, 0x01, 0x11, 0x12, 0x12, 0x22, 0x22,
  db 0x66, 0x69, 0x69, 0x59, 0x58, 0x99, 0x99, 0x88,
  db 0x00, 0x00, 0x00, 0x00, 0xF8, 0xB4, 0x13, 0x51,
  db 0x35, 0x51, 0x51, 0x19, 0x01, 0x00

lut_partial_update:
  db 0x10, 0x18, 0x18, 0x08, 0x18, 0x18, 0x08, 0x00,
  db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  db 0x00, 0x00, 0x00, 0x00, 0x13, 0x14, 0x44, 0x12,
  db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00

digits:
  db 0xfe, 0x82, 0x82, 0x82,  0x82, 0x82, 0x82, 0xfe
  db 0x08, 0x08, 0x08, 0x08,  0x08, 0x08, 0x08, 0x08
  db 0xfe, 0x02, 0x02, 0x02,  0xfe, 0x80, 0x80, 0xfe
  db 0xfe, 0x02, 0x02, 0x02,  0x1e, 0x02, 0x02, 0xfe
  db 0x82, 0x82, 0x82, 0x82,  0xfe, 0x02, 0x02, 0x02
  db 0xfe, 0x80, 0x80, 0x80,  0xfe, 0x02, 0x02, 0xfe
  db 0xfe, 0x80, 0x80, 0x80,  0xfe, 0x82, 0x82, 0xfe
  db 0xfe, 0x02, 0x02, 0x02,  0x02, 0x02, 0x02, 0x02
  db 0xfe, 0x82, 0x82, 0x82,  0xfe, 0x82, 0x82, 0xfe
  db 0xfe, 0x82, 0x82, 0x82,  0xfe, 0x02, 0x02, 0xfe

dec_value:
  db 0x00, 0x00, 0x01, 0x01,  0x02, 0x03, 0x03, 0x04
  db 0x05, 0x05, 0x06, 0x06,  0x07, 0x08, 0x08, 0x09

.org 0xfff2
  dw timer_interrupt       ; Timer_A2 TACCR0, CCIFG
.org 0xfffe
  dw start

